"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.$sqllike = exports.$valuesForKeyPath = exports.TSQualifier = exports.LAZY_INTERSECTION = void 0;
const commons_1 = require("./commons");
const types_1 = require("./types");
const compare_1 = require("./compare");
const tsobject_1 = require("./tsobject");
exports.LAZY_INTERSECTION = false;
class TSQualifier {
    constructor(operator, operands) {
        this.operator = operator;
        this._operands = operands;
    }
    // Tree management qualifiers
    static OR(conds = []) { return new this('OR', [...conds]); }
    static AND(conds = []) { return new this('AND', [...conds]); }
    static NOT(cond) { return new this('NOT', [cond]); }
    // exsitence qualifiers
    static OK(key) { return new this('OK', [_split(key)]); }
    static KO(key) { return new this('KO', [_split(key)]); }
    // key-value(s) qualifiers
    static EQ(key, value) { return new this('EQ', [_split(key), value]); }
    static NEQ(key, value) { return new this('NEQ', [_split(key), value]); }
    static LT(key, value) { return new this('LT', [_split(key), value]); }
    static LTE(key, value) { return new this('LTE', [_split(key), value]); }
    static GT(key, value) { return new this('GT', [_split(key), value]); }
    static GTE(key, value) { return new this('GTE', [_split(key), value]); }
    // like 'SQL LIKE' operator
    static LIKE(key, value) { return new this('LIKE', [_split(key), value]); }
    static IN(key, values) {
        if (values.length === 0) {
            throw 'TSQualifier.IN(): Allways false condition with empty values';
        }
        return values.length === 1 ? this.EQ(key, values[0]) : new this('IN', [_split(key), values]);
    }
    static NIN(key, values) {
        if (values.length === 0) {
            throw 'TSQualifier.NIN(): Allways true condition with empty values';
        }
        return values.length === 1 ? this.NEQ(key, values[0]) : new this('NIN', [_split(key), values]);
    }
    static INCLUDES(key1, key2, value) {
        if (!(0, commons_1.$ok)(value))
            throw 'TSQualifier.INCLUDING(): should have a not null and not undefined comparison value';
        return this.AND([this.LTE(key1, value), this.GT(key2, value)]);
    }
    static INCLUDED(key, value1, value2, canUnspecify = false, strict = exports.LAZY_INTERSECTION) {
        let target = this.intersectionQualifiers(value1, value2, key, key, strict, canUnspecify);
        if (!target.length)
            throw 'TSQualifier.INCLUDED(): Bad values.';
        return target.length === 1 ? target[0] : this.AND(target);
    }
    static INTERSECTS(key1, key2, value1, value2, canUnspecify = false, strict = exports.LAZY_INTERSECTION) {
        let quals = this.intersectionQualifiers(value1, value2, key1, key2, strict, canUnspecify);
        return quals.length === 1 ? quals[0] : this.AND(quals);
    }
    static INRANGE(key, value) {
        let qual = this.AND();
        qual.addRange(key, value);
        if (!qual._operands.length)
            throw 'TSQualifier.INRANGE(): Bad TSRange value';
        return qual._operands.length === 1 ? qual._operands[0] : qual;
    }
    static intersectionQualifiers(A, B, C, D, strict, canUnspecify) {
        let target = [];
        if (!(0, commons_1.$ok)(A) && !(0, commons_1.$ok)(B))
            throw 'TSQualifier._addintersectionQuals() must at least have a research millestone';
        const gt = strict ? 'GT' : 'GTE';
        const lt = strict ? 'LT' : 'LTE';
        if (canUnspecify) {
            if ((0, commons_1.$ok)(A))
                target.push(this.OR([new this(gt, [_split(D), A]), this.EQ(D, null)]));
            if ((0, commons_1.$ok)(B))
                target.push(this.OR([new this(lt, [_split(C), B]), this.EQ(C, null)]));
        }
        else {
            if ((0, commons_1.$ok)(A))
                target.push(new this(gt, [_split(D), A]));
            if ((0, commons_1.$ok)(B))
                target.push(new this(lt, [_split(C), B]));
        }
        return target;
    }
    get isComposite() { return this.operator === 'AND' || this.operator === 'OR' || this.operator === 'NOT'; }
    get isKeyValue() { return !this.isComposite && this.operator !== 'OK' && this.operator !== 'KO'; }
    key() { return this.isComposite ? undefined : this._operands[0].join('.'); }
    keyArray() { return this.isComposite ? [] : this._operands[0]; }
    value() { return this.isKeyValue ? this._operands[1] : undefined; }
    conditions() { return this.isComposite ? this._operands : []; }
    add(cond) {
        if (this.operator !== 'OR' && this.operator !== 'AND') {
            throw 'TSQualifier.prototype.add(): trying to add condition on wrong qualifier';
        }
        if (cond instanceof TSQualifier && cond.operator === this.operator) {
            this._operands = [...this._operands, ...cond._operands];
        }
        else {
            this._operands.push(cond);
        }
    }
    addValue(key, value) {
        const qualifier = this.constructor.EQ(key, value);
        this.add(qualifier);
    }
    addNotValue(key, value) {
        const qualifier = this.constructor.NEQ(key, value);
        this.add(qualifier);
    }
    addRange(key, range) {
        if (!range.isValid || range.isEmpty) {
            throw 'TSQualifier.prototype.addRange(): trying to add an empty or invalid range value';
        }
        if (this.operator === 'AND') {
            if (range.length === 1) {
                this.addValue(key, range.location);
            }
            else {
                this.add(this.constructor.GTE(key, range.location));
                this.add(this.constructor.LT(key, range.maxRange));
            }
        }
        else {
            const qualifier = this.constructor.INRANGE(key, range);
            this.add(qualifier);
        }
    }
    addInclusion(key, value1, value2, canUnspecify = false, strict = exports.LAZY_INTERSECTION) {
        this.addIntersection(key, key, value1, value2, canUnspecify, strict);
    }
    addIntersection(key1, key2, value1, value2, canUnspecify = false, strict = exports.LAZY_INTERSECTION) {
        let quals = this.constructor._intersectionQuals(value1, value2, key1, key2, strict, canUnspecify);
        if (quals.length === 1) {
            this.add(quals[0]);
        }
        else if (this.operator === 'AND') {
            quals.forEach(q => this.add(q));
        }
        else {
            this.add(this.constructor.AND(quals));
        }
    }
    inverse() { return this.constructor.NOT(this); }
    validateValue(v, validateValueForCondition) {
        let vals;
        let op;
        switch (this.operator) {
            case 'AND':
                for (let cond of this._operands) {
                    if (cond instanceof TSQualifier) {
                        if (!cond.validateValue(v))
                            return false;
                    }
                    else if ((0, commons_1.$ok)(validateValueForCondition)) {
                        if (!validateValueForCondition(v, cond))
                            return false;
                    }
                    else {
                        throw 'need a validateValueForCondition() callback to interpret a specific AND condition';
                    }
                }
                return true;
            case 'OR':
                for (let cond of this._operands) {
                    if (cond instanceof TSQualifier) {
                        if (cond.validateValue(v))
                            return true;
                    }
                    else if ((0, commons_1.$ok)(validateValueForCondition)) {
                        if (validateValueForCondition(v, cond))
                            return true;
                    }
                    else {
                        throw 'need a validateValueForCondition() callback to interpret a specific OR condition';
                    }
                }
                return false;
            case 'NOT':
                const cond = this._operands[0];
                if (cond instanceof TSQualifier) {
                    return !cond.validateValue(v);
                }
                else if ((0, commons_1.$ok)(validateValueForCondition)) {
                    return !validateValueForCondition(v, cond);
                }
                throw 'need a validateValueForCondition() callback to interpret a specific NOT condition';
            case 'OK':
                return _valuesForKeyPath(v, this._operands[0]).length > 0;
            case 'KO':
                return _valuesForKeyPath(v, this._operands[0]).length === 0;
            case 'EQ':
                vals = _valuesForKeyPath(v, this._operands[0]);
                op = this._operands[1];
                for (let v of vals) {
                    if ((0, compare_1.$equal)(v, op))
                        return true;
                }
                ;
                return false;
            case 'NEQ':
                vals = _valuesForKeyPath(v, this._operands[0]);
                op = this._operands[1];
                for (let v of vals) {
                    if (!(0, compare_1.$equal)(v, op))
                        return true;
                }
                ;
                return false;
            case 'LT':
                vals = _valuesForKeyPath(v, this._operands[0]);
                op = this._operands[1];
                for (let v of vals) {
                    if ((0, compare_1.$compare)(v, op) === types_1.Ascending)
                        return true;
                }
                ;
                return false;
            case 'LTE':
                vals = _valuesForKeyPath(v, this._operands[0]);
                op = this._operands[1];
                for (let v of vals) {
                    const comp = (0, compare_1.$compare)(v, op);
                    if (comp === types_1.Same || comp === types_1.Ascending)
                        return true;
                }
                return false;
            case 'GT':
                vals = _valuesForKeyPath(v, this._operands[0]);
                op = this._operands[1];
                for (let v of vals) {
                    if ((0, compare_1.$compare)(v, op) === types_1.Descending)
                        return true;
                }
                ;
                return false;
            case 'GTE':
                vals = _valuesForKeyPath(v, this._operands[0]);
                op = this._operands[1];
                for (let v of vals) {
                    const comp = (0, compare_1.$compare)(v, op);
                    if (comp === types_1.Same || comp === types_1.Descending)
                        return true;
                }
                return false;
            case 'LIKE':
                vals = _valuesForKeyPath(v, this._operands[0]);
                op = this._operands[1];
                for (let v of vals) {
                    if ($sqllike(v, op))
                        return true;
                }
                ;
                return false;
            case 'IN':
                vals = _valuesForKeyPath(v, this._operands[0]);
                op = this._operands[1];
                for (let v of vals) {
                    for (let e of op) {
                        if ((0, compare_1.$equal)(e, v))
                            return true;
                    }
                }
                ;
                return false;
            case 'NIN':
                vals = _valuesForKeyPath(v, this._operands[0]);
                op = this._operands[1];
                for (let v of vals) {
                    for (let e of op) {
                        if ((0, compare_1.$equal)(e, v))
                            return false;
                    }
                }
                ;
                return true;
        }
    }
    filterValues(values) {
        const ret = [];
        if ((0, commons_1.$ok)(values)) {
            for (let v of values) {
                if (this.validateValue(v)) {
                    ret.push(v);
                }
            }
        }
        return ret;
    }
}
exports.TSQualifier = TSQualifier;
function $valuesForKeyPath(v, path) { return (0, commons_1.$ok)(v) ? _valuesForKeyPath(v, _split(path)) : v; }
exports.$valuesForKeyPath = $valuesForKeyPath;
function _split(key) { return key.split('.'); }
function _valuesForKeyPath(v, keyPath) {
    let values = [];
    if ((0, commons_1.$ok)(v)) {
        values = [v];
        for (let key of keyPath) {
            let newValues = [];
            for (let v of values) {
                v = v[key];
                if ((0, commons_1.$ok)(v)) {
                    if ((0, tsobject_1.$iscollection)(v)) {
                        v = v.getItems();
                    }
                    if ((0, commons_1.$isarray)(v) || (v instanceof Set)) {
                        for (let o of v) {
                            newValues.push(o);
                        }
                    }
                    else {
                        newValues.push(v);
                    }
                }
            }
            if (newValues.length === 0) {
                return [];
            }
            values = newValues;
        }
    }
    return values;
}
if (!('sqlLike' in String.prototype)) {
    // we mimic SQL Like operator here
    String.prototype.sqlLike = function sqlLike(like) {
        if (!(0, commons_1.$ok)(like)) {
            return false;
        }
        like = like.replace(new RegExp("([\\.\\\\\\+\\*\\?\\[\\^\\]\\$\\(\\)\\{\\}\\=\\!\\<\\>\\|\\:\\-])", "g"), "\\$1");
        like = like.replace(/%/g, '.*').replace(/_/g, '.');
        return RegExp('^' + like + '$', 'gi').test(this);
    };
}
function $sqllike(a, b) {
    if (!(0, commons_1.$ok)(a))
        return false;
    return (0, commons_1.$isstring)(a) ? a.sqlLike(b) : `${a}`.sqlLike(b); // if a as no specific toString() method, this will return inconsistent result
}
exports.$sqllike = $sqllike;
if (!('filterWithQualifier' in Array.prototype)) {
    Array.prototype.filterWithQualifier = function filterWithQualifier(qual) { return qual.filterValues(this); };
}
//# sourceMappingURL=tsqualifier.js.map